@@ -1,4 +1,4 @@
-# Copyright 2004-2015 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -26,11 +26,15 @@
 # When updating this file, consider if lint.py or warp.py also need
 # updating.
 
-import renpy.display
+from __future__ import print_function
 
+import renpy.display
+import renpy.test
+
+import hashlib
 import re
 import time
-import md5
+
 
 def statement_name(name):
     """
@@ -40,6 +44,7 @@ def statement_name(name):
     for i in renpy.config.statement_callbacks:
         i(name)
 
+
 def next_node(n):
     """
     Indicates the next node that should be executed. When a statement
@@ -49,11 +54,13 @@ def next_node(n):
 
     renpy.game.context().next_node = n
 
+
 class ParameterInfo(object):
     """
     This class is used to store information about parameters to a
     label.
     """
+
     def __init__(self, parameters, positional, extrapos, extrakw):
 
         # A list of parameter name, default value pairs.
@@ -71,7 +78,6 @@ class ParameterInfo(object):
         # any. None if no such variable exists.
         self.extrakw = extrakw
 
-
     def apply(self, args, kwargs, ignore_errors=False):
         """
         Applies `args` and `kwargs` to these parameters. Returns
@@ -92,7 +98,6 @@ class ParameterInfo(object):
         if kwargs is None:
             kwargs = { }
 
-
         for name, value in zip(self.positional, args):
             if name in values:
                 if not ignore_errors:
@@ -137,6 +142,7 @@ class ParameterInfo(object):
 
         return rv
 
+
 def apply_arguments(parameters, args, kwargs, ignore_errors=False):
 
     if parameters is None:
@@ -187,12 +193,37 @@ class ArgumentInfo(object):
 
         return tuple(args), kwargs
 
+    def get_code(self):
+
+        l = [ ]
+
+        for keyword, expression in self.arguments:
+            if keyword is not None:
+                l.append("{}={}".format(keyword, expression))
+            else:
+                l.append(expression)
+
+        if self.extrapos is not None:
+            l.append("*" + self.extrapos)
+
+        if self.extrakw is not None:
+            l.append("**" + self.extrakw)
+
+        return "(" + ", ".join(l) + ")"
+
 
 def __newobj__(cls, *args):
     return cls.__new__(cls, *args)
 
-# This represents a string containing python code.
+
+# A list of pyexprs that need to be precompiled.
+pyexpr_list = [ ]
+
+
 class PyExpr(unicode):
+    """
+    Represents a string containing python code.
+    """
 
     __slots__ = [
         'filename',
@@ -204,10 +235,15 @@ class PyExpr(unicode):
         self.filename = filename
         self.linenumber = linenumber
 
+        # Queue the string for precompilation.
+        if self and (renpy.game.script.all_pyexpr is not None):
+            renpy.game.script.all_pyexpr.append(self)
+
         return self
 
     def __getnewargs__(self):
-        return (unicode(self), self.filename, self.linenumber) # E1101
+        return (unicode(self), self.filename, self.linenumber)  # E1101
+
 
 class PyCode(object):
 
@@ -258,14 +294,14 @@ class PyCode(object):
             pass
 
         code = self.source
-        if isinstance(code, renpy.python.ast.AST): #@UndefinedVariable
-            code = renpy.python.ast.dump(code) #@UndefinedVariable
+        if isinstance(code, renpy.python.ast.AST):  # @UndefinedVariable
+            code = renpy.python.ast.dump(code)  # @UndefinedVariable
 
-        self.hash = chr(renpy.bytecode_version) + md5.md5(repr(self.location) + code.encode("utf-8")).digest()
+        self.hash = chr(renpy.bytecode_version) + hashlib.md5(repr(self.location) + code.encode("utf-8")).digest()
         return self.hash
 
 
-def chain_block(block, next): #@ReservedAssignment
+def chain_block(block, next):  # @ReservedAssignment
     """
     This is called to chain together all of the nodes in a block. Node
     n is chained with node n+1, while the last node is chained with
@@ -291,11 +327,14 @@ class Scry(object):
     def __getattr__(self, name):
         return None
 
-    def next(self): #@ReservedAssignment
+    def next(self):  # @ReservedAssignment
         if self._next is None:
             return None
         else:
-            return self._next.scry()
+            try:
+                return self._next.scry()
+            except:
+                return None
 
 
 class Node(object):
@@ -323,6 +362,13 @@ class Node(object):
     # take_translations.
     translation_relevant = False
 
+    # How does the node participate in rollback?
+    #
+    # * "normal" in normal mode.
+    # * "never" generally never.
+    # * "force" force it to start.
+    rollback = "normal"
+
     def __init__(self, loc):
         """
         Initializes this Node object.
@@ -331,7 +377,7 @@ class Node(object):
         logical line on which this Node node starts.
         """
 
-        self.filename, self.linenumber  = loc
+        self.filename, self.linenumber = loc
         self.name = None
         self.next = None
 
@@ -365,7 +411,7 @@ class Node(object):
     # get_init is only present on statements that define it.
     get_init = None
 
-    def chain(self, next): #@ReservedAssignment
+    def chain(self, next):  # @ReservedAssignment
         """
         This is called with the Node node that should be followed after
         executing this node, and all nodes that this node
@@ -421,7 +467,7 @@ class Node(object):
         """
 
         rv = Scry()
-        rv._next = self.next # W0201
+        rv._next = self.next  # W0201
         return rv
 
     def restructure(self, callback):
@@ -456,6 +502,16 @@ class Node(object):
         # Does nothing by default.
         return
 
+    warp = False
+
+    def can_warp(self):
+        """
+        Returns true if this should be run while warping, False otherwise.
+        """
+
+        return self.warp
+
+
 def say_menu_with(expression, callback):
     """
     This handles the with clause of a say or menu statement.
@@ -475,6 +531,10 @@ def say_menu_with(expression, callback):
         # renpy.game.interface.set_transition(what)
         callback(what)
 
+
+fast_who_pattern = re.compile(r'[a-zA-Z_][a-zA-Z_0-9]*$')
+
+
 def eval_who(who, fast=None):
     """
     Evaluates the `who` parameter to a say statement.
@@ -484,7 +544,7 @@ def eval_who(who, fast=None):
         return None
 
     if fast is None:
-        fast = bool(re.match(renpy.parser.word_regexp + "$", who))
+        fast = bool(fast_who_pattern.match(who))
 
     if fast:
 
@@ -503,6 +563,7 @@ def eval_who(who, fast=None):
 
     return renpy.python.py_eval(who)
 
+
 class Say(Node):
 
     __slots__ = [
@@ -512,6 +573,7 @@ class Say(Node):
         'with_',
         'interact',
         'attributes',
+        'arguments',
         ]
 
     def diff_info(self):
@@ -521,9 +583,10 @@ class Say(Node):
         self = Node.__new__(cls)
         self.attributes = None
         self.interact = True
+        self.arguments = None
         return self
 
-    def __init__(self, loc, who, what, with_, interact=True, attributes=None):
+    def __init__(self, loc, who, what, with_, interact=True, attributes=None, arguments=None):
 
         super(Say, self).__init__(loc)
 
@@ -542,6 +605,7 @@ class Say(Node):
         self.what = what
         self.with_ = with_
         self.interact = interact
+        self.arguments = arguments
 
         # A tuple of attributes that are applied to the character that's
         # speaking, or None to disable this behavior.
@@ -569,6 +633,9 @@ class Say(Node):
             rv.append("with")
             rv.append(self.with_)
 
+        if self.arguments:
+            rv.append(self.arguments.get_code())
+
         return " ".join(rv)
 
     def execute(self):
@@ -583,29 +650,36 @@ class Say(Node):
             who = eval_who(self.who, self.who_fast)
 
             if not (
-                (who is None) or
-                callable(who) or
-                isinstance(who, basestring) ):
+                    (who is None) or
+                    callable(who) or
+                    isinstance(who, basestring) ):
 
                 raise Exception("Sayer %s is not a function or string." % self.who.encode("utf-8"))
 
             what = self.what
             if renpy.config.say_menu_text_filter:
-                what = renpy.config.say_menu_text_filter(what) # E1102
+                what = renpy.config.say_menu_text_filter(what)  # E1102
 
             renpy.store._last_raw_what = what
 
+            if self.arguments is not None:
+                args, kwargs = self.arguments.evaluate()
+            else:
+                args = tuple()
+                kwargs = dict()
+
             if getattr(who, "record_say", True):
                 renpy.store._last_say_who = self.who
                 renpy.store._last_say_what = what
+                renpy.store._last_say_args = args
+                renpy.store._last_say_kwargs = kwargs
 
             say_menu_with(self.with_, renpy.game.interface.set_transition)
-            renpy.exports.say(who, what, interact=self.interact)
+            renpy.exports.say(who, what, interact=self.interact, *args, **kwargs)
 
         finally:
             renpy.game.context().say_attributes = None
 
-
     def predict(self):
 
         old_attributes = renpy.game.context().say_attributes
@@ -635,13 +709,7 @@ class Say(Node):
     def scry(self):
         rv = Node.scry(self)
 
-        if self.who is not None:
-            if self.who_fast:
-                who = getattr(renpy.store, self.who)
-            else:
-                who = renpy.python.py_eval(self.who)
-        else:
-            who = None
+        who = eval_who(self.who, self.who_fast)
 
         if self.interact:
             renpy.exports.scry_say(who, rv)
@@ -650,8 +718,10 @@ class Say(Node):
 
         return rv
 
+
 # Copy the descriptor.
-setattr(Say, "with", Say.with_) # E1101
+setattr(Say, "with", Say.with_)  # E1101
+
 
 class Init(Node):
 
@@ -666,7 +736,6 @@ class Init(Node):
         self.block = block
         self.priority = priority
 
-
     def get_children(self, f):
         f(self)
 
@@ -679,13 +748,14 @@ class Init(Node):
     # We handle chaining specially. We want to chain together the nodes in
     # the block, but we want that chain to end in None, and we also want
     # this node to just continue on to the next node in normal execution.
-    def chain(self, next): #@ReservedAssignment
+    def chain(self, next):  # @ReservedAssignment
         self.next = next
 
         chain_block(self.block, None)
 
     def execute(self):
         next_node(self.next)
+        renpy.execution.not_infinite_loop(60)
         statement_name("init")
 
     def restructure(self, callback):
@@ -694,6 +764,8 @@ class Init(Node):
 
 class Label(Node):
 
+    rollback = "force"
+
     translation_relevant = True
 
     __slots__ = [
@@ -734,7 +806,7 @@ class Label(Node):
         for i in self.block:
             i.get_children(f)
 
-    def chain(self, next): #@ReservedAssignment
+    def chain(self, next):  # @ReservedAssignment
 
         if self.block:
             self.next = self.block[0]
@@ -788,7 +860,12 @@ class Python(Node):
         super(Python, self).__init__(loc)
 
         self.hide = hide
-        self.code = PyCode(python_code, loc=loc, mode='exec')
+
+        if hide:
+            self.code = PyCode(python_code, loc=loc, mode='hide')
+        else:
+            self.code = PyCode(python_code, loc=loc, mode='exec')
+
         self.store = store
 
     def diff_info(self):
@@ -814,6 +891,7 @@ class Python(Node):
         rv.interacts = True
         return rv
 
+
 class EarlyPython(Node):
 
     __slots__ = [
@@ -838,7 +916,12 @@ class EarlyPython(Node):
         super(EarlyPython, self).__init__(loc)
 
         self.hide = hide
-        self.code = PyCode(python_code, loc=loc, mode='exec')
+
+        if hide:
+            self.code = PyCode(python_code, loc=loc, mode='hide')
+        else:
+            self.code = PyCode(python_code, loc=loc, mode='exec')
+
         self.store = store
 
     def diff_info(self):
@@ -846,6 +929,7 @@ class EarlyPython(Node):
 
     def execute(self):
         next_node(self.next)
+        renpy.execution.not_infinite_loop(60)
         statement_name("python early")
 
     def early_execute(self):
@@ -854,6 +938,7 @@ class EarlyPython(Node):
         if self.code.bytecode:
             renpy.python.py_exec_bytecode(self.code.bytecode, self.hide, store=self.store)
 
+
 class Image(Node):
 
     __slots__ = [
@@ -1024,7 +1109,7 @@ def show_imspec(imspec, atl=None):
 
     at_list = [ renpy.python.py_eval(i) for i in at_list ]
 
-    layer = renpy.exports.default_layer(layer, tag or name, expression)
+    layer = renpy.exports.default_layer(layer, tag or name, expression and (tag is None))
 
     renpy.config.show(name,
                       at_list=at_list,
@@ -1043,6 +1128,8 @@ class Show(Node):
         'atl',
         ]
 
+    warp = True
+
     def __init__(self, loc, imspec, atl=None):
         """
         @param imspec: A triple consisting of an image name (itself a
@@ -1074,6 +1161,8 @@ class Show(Node):
 
 class ShowLayer(Node):
 
+    warp = True
+
     __slots__ = [
         'layer',
         'at_list',
@@ -1118,6 +1207,8 @@ class Scene(Node):
         'atl',
         ]
 
+    warp = True
+
     def __init__(self, loc, imgspec, layer, atl=None):
         """
         @param imspec: A triple consisting of an image name (itself a
@@ -1168,6 +1259,8 @@ class Hide(Node):
         'imspec',
         ]
 
+    warp = True
+
     def __init__(self, loc, imgspec):
         """
         @param imspec: A triple consisting of an image name (itself a
@@ -1196,7 +1289,6 @@ class Hide(Node):
         elif len(self.imspec) == 7:
             name, _expression, tag, _at_list, layer, _zorder, _behind = self.imspec
 
-
         if tag is None:
             tag = name[0]
 
@@ -1275,7 +1367,6 @@ class With(Node):
         except:
             pass
 
-
         return [ self.next ]
 
 
@@ -1351,7 +1442,7 @@ class Return(Node):
         return (Return, )
 
     # We don't care what the next node is.
-    def chain(self, next): #@ReservedAssignment
+    def chain(self, next):  # @ReservedAssignment
         self.next = None
         return
 
@@ -1364,6 +1455,16 @@ class Return(Node):
         else:
             renpy.store._return = None
 
+        ctx = renpy.game.context()
+
+        if renpy.game.context().init_phase:
+            if len(ctx.return_stack) == 0:
+
+                if renpy.config.developer:
+                    raise Exception("Unexpected return during the init phase.")
+
+                return
+
         next_node(renpy.game.context().lookup_return(pop=True))
         renpy.game.context().pop_dynamic()
 
@@ -1385,14 +1486,21 @@ class Menu(Node):
         'items',
         'set',
         'with_',
+        'has_caption',
         ]
 
-    def __init__(self, loc, items, set, with_): #@ReservedAssignment
+    def __new__(cls, *args, **kwargs):
+        self = Node.__new__(cls)
+        self.has_caption = False
+        return self
+
+    def __init__(self, loc, items, set, with_, has_caption):  # @ReservedAssignment
         super(Menu, self).__init__(loc)
 
         self.items = items
         self.set = set
         self.with_ = with_
+        self.has_caption = has_caption
 
     def diff_info(self):
         return (Menu,)
@@ -1406,7 +1514,7 @@ class Menu(Node):
                     i.get_children(f)
 
     # Blocks of statements in a choice continue after the menu.
-    def chain(self, next): #@ReservedAssignment
+    def chain(self, next):  # @ReservedAssignment
 
         self.next = next
 
@@ -1424,7 +1532,11 @@ class Menu(Node):
     def execute(self):
 
         next_node(self.next)
-        statement_name("menu")
+
+        if self.has_caption:
+            statement_name("menu-with-caption")
+        else:
+            statement_name("menu")
 
         choices = [ ]
         narration = [ ]
@@ -1454,7 +1566,6 @@ class Menu(Node):
         else:
             next_node(self.next)
 
-
     def predict(self):
         rv = [ ]
 
@@ -1482,7 +1593,8 @@ class Menu(Node):
             if block is not None:
                 callback(block)
 
-setattr(Menu, "with", Menu.with_) # E1101
+
+setattr(Menu, "with", Menu.with_)  # E1101
 
 
 # Goto is considered harmful. So we decided to name it "jump"
@@ -1494,7 +1606,7 @@ class Jump(Node):
         'expression',
         ]
 
-    def  __init__(self, loc, target, expression):
+    def __init__(self, loc, target, expression):
         super(Jump, self).__init__(loc)
 
         self.target = target
@@ -1504,7 +1616,7 @@ class Jump(Node):
         return (Jump, self.target, self.expression)
 
     # We don't care what our next node is.
-    def chain(self, next): #@ReservedAssignment
+    def chain(self, next):  # @ReservedAssignment
         self.next = None
         return
 
@@ -1573,7 +1685,7 @@ class While(Node):
         for i in self.block:
             i.get_children(f)
 
-    def chain(self, next): #@ReservedAssignment
+    def chain(self, next):  # @ReservedAssignment
         self.next = next
         chain_block(self.block, self)
 
@@ -1602,6 +1714,7 @@ class While(Node):
     def restructure(self, callback):
         callback(self.block)
 
+
 class If(Node):
 
     __slots__ = [ 'entries' ]
@@ -1625,7 +1738,7 @@ class If(Node):
             for i in block:
                 i.get_children(f)
 
-    def chain(self, next): #@ReservedAssignment
+    def chain(self, next):  # @ReservedAssignment
         self.next = next
 
         for _condition, block in self.entries:
@@ -1662,17 +1775,21 @@ class If(Node):
         for _condition, block in self.entries:
             callback(block)
 
+
 class UserStatement(Node):
 
     __slots__ = [
         'line',
         'parsed',
         'block',
-        'translatable' ]
+        'translatable',
+        'code_block',
+        ]
 
     def __new__(cls, *args, **kwargs):
         self = Node.__new__(cls)
         self.block = [ ]
+        self.code_block = None
         self.translatable = False
         return self
 
@@ -1681,14 +1798,60 @@ class UserStatement(Node):
         super(UserStatement, self).__init__(loc)
         self.line = line
         self.block = block
+        self.code_block = None
         self.parsed = None
 
-        # Do not store the parse quite yet.
-        _parse_info = renpy.statements.parse(self, self.line, self.block)
+        self.name = self.call("label")
+
+        # Do not store the parse.
+        self.parsed = None
+
+    def __repr__(self):
+        return "<UserStatement {!r}>".format(self.line)
+
+    def get_children(self, f):
+        f(self)
+
+        if not self.code_block:
+            return
+
+        for i in self.code_block:
+            i.get_children(f)
+
+    def chain(self, next):  # @ReservedAssignment
+        self.next = next
+
+        if self.code_block is not None:
+            chain_block(self.code_block, next)
+
+    def replace_next(self, old, new):
+        Node.replace_next(self, old, new)
+
+        if (self.code_block) and (self.code_block[0] is old):
+            self.code_block.insert(0, new)
+
+    def restructure(self, callback):
+        if self.code_block:
+            callback(self.code_block)
 
     def diff_info(self):
         return (UserStatement, self.line)
 
+    def call(self, method, *args, **kwargs):
+
+        parsed = self.parsed
+        if parsed is None:
+            parsed = renpy.statements.parse(self, self.line, self.block)
+            self.parsed = parsed
+
+        return renpy.statements.call(method, parsed, *args, **kwargs)
+
+    def execute_init(self):
+        self.call("execute_init")
+
+    def get_init(self):
+        return 0, self.execute_init
+
     def execute(self):
         next_node(self.get_next())
         statement_name(self.get_name())
@@ -1704,15 +1867,6 @@ class UserStatement(Node):
 
         return [ self.get_next() ]
 
-    def call(self, method, *args, **kwargs):
-
-        parsed = self.parsed
-        if parsed is None:
-            parsed = renpy.statements.parse(self, self.line, self.block)
-            self.parsed = parsed
-
-        return renpy.statements.call(method, parsed, *args, **kwargs)
-
     def get_name(self):
         parsed = self.parsed
         if parsed is None:
@@ -1722,7 +1876,12 @@ class UserStatement(Node):
         return renpy.statements.get_name(parsed)
 
     def get_next(self):
-        rv = self.call("next")
+
+        if self.code_block and len(self.code_block):
+            rv = self.call("next", self.code_block[0].name)
+        else:
+            rv = self.call("next")
+
         if rv is not None:
             return renpy.game.script.lookup(rv)
         else:
@@ -1737,17 +1896,28 @@ class UserStatement(Node):
     def get_code(self, dialogue_filter=None):
         return self.line
 
+    def can_warp(self):
+
+        if self.call("warp"):
+            return True
+
+        return False
+
+
 def create_store(name):
     if name not in renpy.config.special_namespaces:
         renpy.python.create_store(name)
 
+
 class StoreNamespace(object):
+
     def __init__(self, store):
         self.store = store
 
     def set(self, name, value):
         renpy.python.store_dicts[self.store][name] = value
 
+
 def get_namespace(store):
     """
     Returns the namespace object for `store`, and a flag that is true if the
@@ -1759,6 +1929,14 @@ def get_namespace(store):
 
     return StoreNamespace(store), False
 
+
+# Config variables that are set twice - once when the rpy is first loaded,
+# and then again at init time.
+EARLY_CONFIG = { "save_directory" }
+
+define_statements = [ ]
+
+
 class Define(Node):
 
     __slots__ = [
@@ -1785,11 +1963,17 @@ class Define(Node):
     def early_execute(self):
         create_store(self.store)
 
+        if self.store == "store.config" and self.varname in EARLY_CONFIG:
+            value = renpy.python.py_eval_bytecode(self.code.bytecode)
+            setattr(renpy.config, self.varname, value)
+
     def execute(self):
 
         next_node(self.next)
         statement_name("define")
 
+        define_statements.append(self)
+
         value = renpy.python.py_eval_bytecode(self.code.bytecode)
 
         if self.store == 'store':
@@ -1801,10 +1985,30 @@ class Define(Node):
         ns, _special = get_namespace(self.store)
         ns.set(self.varname, value)
 
+    def redefine(self, stores):
+
+        if self.store not in stores:
+            return
+
+        value = renpy.python.py_eval_bytecode(self.code.bytecode)
+        ns, _special = get_namespace(self.store)
+        ns.set(self.varname, value)
+
+
+def redefine(stores):
+    """
+    Re-runs the given define statements.
+    """
+
+    for i in define_statements:
+        i.redefine(stores)
+
+
 
 # All the default statements, in the order they were registered.
 default_statements = [ ]
 
+
 class Default(Node):
 
     __slots__ = [
@@ -1837,12 +2041,15 @@ class Default(Node):
         next_node(self.next)
         statement_name("default")
 
-        default_statements.append(self)
-
-        _ns, special = get_namespace(self.store)
+        ns, special = get_namespace(self.store)
 
         if special:
-            raise Exception("The default statement can't be used with the special namespace %r." % self.store)
+
+            value = renpy.python.py_eval_bytecode(self.code.bytecode)
+            ns.set_default(self.varname, value)
+            return
+
+        default_statements.append(self)
 
         if self.store == 'store':
             renpy.dump.definitions.append((self.varname, self.filename, self.linenumber))
@@ -1853,13 +2060,22 @@ class Default(Node):
         d = renpy.python.store_dicts[self.store]
 
         defaults_set = d.get("_defaults_set", None)
+
         if defaults_set is None:
             d["_defaults_set"] = defaults_set = renpy.python.RevertableSet()
+            d.ever_been_changed.add("_defaults_set")
 
         if self.varname not in defaults_set:
-            d[self.varname] = renpy.python.py_eval_bytecode(self.code.bytecode)
+
+            if start or (self.varname not in d.ever_been_changed):
+                d[self.varname] = renpy.python.py_eval_bytecode(self.code.bytecode)
+
+            d.ever_been_changed.add(self.varname)
+
             defaults_set.add(self.varname)
+
         else:
+
             if start and renpy.config.developer:
                 raise Exception("{}.{} is being given a default a second time.".format(self.store, self.varname))
 
@@ -1903,7 +2119,7 @@ class Screen(Node):
 class Translate(Node):
     """
     A translation block, produced either by explicit translation statements
-    or implicit translation blocs.
+    or implicit translation blocks.
 
     If language is None, when executed this transfers control to the translate
     statement in the current language, if any, and otherwise runs the block.
@@ -1914,26 +2130,30 @@ class Translate(Node):
     goes to the end of the translate statement in the None language.
     """
 
+    rollback = "never"
+
     translation_relevant = True
 
     __slots__ = [
         "identifier",
+        "alternate",
         "language",
         "block",
         "after",
         ]
 
-    def __init__(self, loc, identifier, language, block):
+    def __init__(self, loc, identifier, language, block, alternate=None):
         super(Translate, self).__init__(loc)
 
         self.identifier = identifier
+        self.alternate = alternate
         self.language = language
         self.block = block
 
     def diff_info(self):
         return (Translate, self.identifier, self.language)
 
-    def chain(self, next): #@ReservedAssignment
+    def chain(self, next):  # @ReservedAssignment
         if self.block:
             self.next = self.block[0]
             chain_block(self.block, next)
@@ -1951,6 +2171,9 @@ class Translate(Node):
         if self.after is old:
             self.after = new
 
+    def lookup(self):
+        return renpy.game.script.translator.lookup_translate(self.identifier, getattr(self, "alternate", None))
+
     def execute(self):
 
         statement_name("translate")
@@ -1959,22 +2182,23 @@ class Translate(Node):
             next_node(self.next)
             raise Exception("Translation nodes cannot be run directly.")
 
-        if self.identifier not in renpy.game.persistent._seen_translates: # @UndefinedVariable
-            renpy.game.persistent._seen_translates.add(self.identifier) # @UndefinedVariable
+        if self.identifier not in renpy.game.persistent._seen_translates:  # @UndefinedVariable
+            renpy.game.persistent._seen_translates.add(self.identifier)  # @UndefinedVariable
             renpy.game.seen_translates_count += 1
+            renpy.game.new_translates_count += 1
 
-        next_node(renpy.game.script.translator.lookup_translate(self.identifier))
+        next_node(self.lookup())
 
         renpy.game.context().translate_identifier = self.identifier
-        renpy.game.context().translate_block_language = self.language
+        renpy.game.context().alternate_translate_identifier = getattr(self, "alternate", None)
 
     def predict(self):
-        node = renpy.game.script.translator.lookup_translate(self.identifier)
+        node = self.lookup()
         return [ node ]
 
     def scry(self):
         rv = Scry()
-        rv._next = renpy.game.script.translator.lookup_translate(self.identifier)
+        rv._next = self.lookup()
         return rv
 
     def get_children(self, f):
@@ -1993,6 +2217,8 @@ class EndTranslate(Node):
     resetting the translation identifier.
     """
 
+    rollback = "never"
+
     def __init__(self, loc):
         super(EndTranslate, self).__init__(loc)
 
@@ -2004,7 +2230,7 @@ class EndTranslate(Node):
         statement_name("end translate")
 
         renpy.game.context().translate_identifier = None
-        renpy.game.context().translate_block_language = None
+        renpy.game.context().alternate_translate_identifier = None
 
 
 class TranslateString(Node):
@@ -2017,14 +2243,17 @@ class TranslateString(Node):
     __slots__ = [
         "language",
         "old",
-        "new"
+        "new",
+        "newloc",
         ]
 
-    def __init__(self, loc, language, old, new):
+    def __init__(self, loc, language, old, new, newloc):
         super(TranslateString, self).__init__(loc)
         self.language = language
+
         self.old = old
         self.new = new
+        self.newloc = newloc
 
     def diff_info(self):
         return (TranslateString,)
@@ -2033,7 +2262,9 @@ class TranslateString(Node):
         next_node(self.next)
         statement_name("translate string")
 
-        renpy.translation.add_string_translation(self.language, self.old, self.new)
+        newloc = getattr(self, "newloc", (self.filename, self.linenumber + 1))
+        renpy.translation.add_string_translation(self.language, self.old, self.new, newloc)
+
 
 class TranslatePython(Node):
     """
@@ -2101,7 +2332,7 @@ class TranslateBlock(Node):
     # We handle chaining specially. We want to chain together the nodes in
     # the block, but we want that chain to end in None, and we also want
     # this node to just continue on to the next node in normal execution.
-    def chain(self, next): #@ReservedAssignment
+    def chain(self, next):  # @ReservedAssignment
         self.next = next
         chain_block(self.block, None)
 
@@ -2112,6 +2343,14 @@ class TranslateBlock(Node):
     def restructure(self, callback):
         callback(self.block)
 
+
+class TranslateEarlyBlock(TranslateBlock):
+    """
+    This is similar to the TranslateBlock, except it runs before deferred
+    styles do.
+    """
+
+
 class Style(Node):
 
     __slots__ = [
@@ -2155,16 +2394,13 @@ class Style(Node):
     def diff_info(self):
         return (Style, self.style_name)
 
-    def execute(self):
-        next_node(self.next)
-        statement_name("style")
-
+    def apply(self):
         if self.variant is not None:
             variant = renpy.python.py_eval(self.variant)
             if not renpy.exports.variant(variant):
                 return
 
-        s = renpy.style.get_or_create_style(self.style_name) # @UndefinedVariable
+        s = renpy.style.get_or_create_style(self.style_name)  # @UndefinedVariable
 
         if self.clear:
             s.clear()
@@ -2180,7 +2416,47 @@ class Style(Node):
 
         if self.properties:
             properties = { }
+
             for name, expr in self.properties.items():
-                properties[name] = renpy.python.py_eval(expr)
+
+                value = renpy.python.py_eval(expr)
+
+                if name == "properties":
+                    properties.update(value)
+                else:
+                    properties[name] = value
 
             s.add_properties(properties)
+
+    def execute(self):
+        next_node(self.next)
+        statement_name("style")
+
+        if renpy.config.defer_styles and renpy.game.context().init_phase:
+            renpy.translation.deferred_styles.append(self)
+            return
+
+        self.apply()
+
+
+class Testcase(Node):
+
+    __slots__ = [
+        'label',
+        'test',
+        ]
+
+    def __init__(self, loc, label, test):
+        super(Testcase, self).__init__(loc)
+
+        self.label = label
+        self.test = test
+
+    def diff_info(self):
+        return (Testcase, self.label)
+
+    def execute(self):
+        next_node(self.next)
+        statement_name("testcase")
+
+        renpy.test.testexecution.testcases[self.label] = self.test
