@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -22,12 +22,13 @@
 # This file contains code that is responsible for storing and executing a
 # Ren'Py script.
 
+from __future__ import print_function
 import renpy
 
+import hashlib
 import os
 import imp
 import difflib
-import md5
 import time
 import marshal
 import struct
@@ -51,12 +52,14 @@ RPYC2_HEADER = "RENPY RPC2"
 # A string
 BYTECODE_FILE = "cache/bytecode.rpyb"
 
+
 class ScriptError(Exception):
     """
     Exception that is raised if the script is somehow inconsistent,
     or otherwise wrong.
     """
 
+
 def collapse_stmts(stmts):
     """
     Returns a flat list containing every statement in the tree
@@ -70,6 +73,7 @@ def collapse_stmts(stmts):
 
     return rv
 
+
 class Script(object):
     """
     This class represents a Ren'Py script, which is parsed out of a
@@ -110,6 +114,7 @@ class Script(object):
         self.namemap = { }
         self.all_stmts = [ ]
         self.all_pycode = [ ]
+        self.all_pyexpr = [ ]
 
         # A list of statements that haven't been analyzed.
         self.need_analysis = [ ]
@@ -122,19 +127,21 @@ class Script(object):
         self.bytecode_dirty = False
 
         self.translator = renpy.translation.ScriptTranslator()
-
         self.init_bytecode()
+
         self.scan_script_files()
 
         self.translator.chain_translates()
 
         self.serial = 0
 
-        self.digest = md5.md5(renpy.version_only)
+        self.digest = hashlib.md5(renpy.version_only)
 
         self.loaded_rpy = False
         self.backup_list = [ ]
 
+        self.duplicate_labels = [ ]
+
     def choose_backupdir(self):
 
         if renpy.mobile:
@@ -145,13 +152,14 @@ class Script(object):
                 return None
 
         import __main__
-        backups = __main__.path_to_saves(renpy.config.gamedir, "backups") # @UndefinedVariable
+        backups = __main__.path_to_saves(renpy.config.gamedir, "backups")  # @UndefinedVariable
 
         if backups is None:
             return
 
         basename = os.path.basename(renpy.config.basedir)
-        backupdir = os.path.join(backups, renpy.exports.fsencode(basename))
+        backupdir = renpy.os.path.join(renpy.exports.fsencode(backups),
+                                       renpy.exports.fsencode(basename))
 
         renpy.exports.write_log("Backing up script files to %r:", backupdir)
 
@@ -195,7 +203,7 @@ class Script(object):
                 continue
 
             try:
-                os.makedirs(os.path.dirname(target_fn), 0700)
+                os.makedirs(os.path.dirname(target_fn), 0o700)
             except:
                 pass
 
@@ -204,7 +212,6 @@ class Script(object):
             except:
                 pass
 
-
     def scan_script_files(self):
         """
         Scan the directories for script files.
@@ -220,7 +227,7 @@ class Script(object):
         # Similar, but for modules:
         self.module_files = [ ]
 
-        for dir, fn in dirlist: #@ReservedAssignment
+        for dir, fn in dirlist:  # @ReservedAssignment
 
             if fn.endswith(".rpy"):
                 if dir is None:
@@ -255,7 +262,7 @@ class Script(object):
 
         initcode = [ ]
 
-        for fn, dir in script_files: #@ReservedAssignment
+        for fn, dir in script_files:  # @ReservedAssignment
             self.load_appropriate_file(".rpyc", ".rpy", dir, fn, initcode)
 
         # Make the sort stable.
@@ -266,10 +273,11 @@ class Script(object):
 
         self.initcode = [ (prio, code) for prio, index, code in initcode ]
 
+        self.translator.chain_translates()
 
     def load_module(self, name):
 
-        files = [ (fn, dir) for fn, dir in self.module_files if fn == name ] #@ReservedAssignment
+        files = [ (fn, dir) for fn, dir in self.module_files if fn == name ]  # @ReservedAssignment
 
         if not files:
             raise Exception("Module %s could not be loaded." % name)
@@ -277,7 +285,7 @@ class Script(object):
         if len(files) > 2:
             raise Exception("Module %s ambiguous, multiple variants exist." % name)
 
-        fn, dir = files[0] #@ReservedAssignment
+        fn, dir = files[0]  # @ReservedAssignment
         initcode = [ ]
 
         self.load_appropriate_file(".rpymc", ".rpym", dir, fn, initcode)
@@ -301,7 +309,6 @@ class Script(object):
                 s.name = (fn, version, self.serial)
                 self.serial += 1
 
-
     def merge_names(self, old_stmts, new_stmts):
 
         old_stmts = collapse_stmts(old_stmts)
@@ -388,7 +395,10 @@ class Script(object):
             filename = renpy.parser.elide_filename(filename)
 
             if not all_stmts[0].filename.lower().endswith(filename.lower()):
-                filename += "c"
+
+                if filename[-1] != "c":
+                    filename += "c"
+
                 for i in all_stmts:
                     i.filename = filename
 
@@ -402,16 +412,25 @@ class Script(object):
                 name = node.name
 
                 if name in self.namemap:
-                    if not isinstance(bad_name, basestring):
-                        bad_name = name
-                        bad_node = node
-                        old_node = self.namemap[name]
 
-            if bad_name is not None:
-                raise ScriptError("Name %s is defined twice, at %s:%d and %s:%d." %
-                                  (repr(bad_name),
-                                   old_node.filename, old_node.linenumber,
-                                   bad_node.filename, bad_node.linenumber))
+                    bad_name = name
+                    bad_node = node
+                    old_node = self.namemap[name]
+
+                    if not isinstance(bad_name, basestring):
+
+                        raise ScriptError("Name %s is defined twice, at %s:%d and %s:%d." %
+                                          (repr(bad_name),
+                                           old_node.filename, old_node.linenumber,
+                                           bad_node.filename, bad_node.linenumber))
+
+                    else:
+                        self.duplicate_labels.append(
+                            u'The label {} is defined twice, at\n  File "{}", line {} and\n  File "{}", line {}.'.format(
+                                bad_name, old_node.filename, old_node.linenumber, bad_node.filename, bad_node.linenumber))
+
+                # Add twice, so we can find duplicates in the same file.
+                self.namemap[name] = node
 
         self.update_bytecode()
 
@@ -448,7 +467,6 @@ class Script(object):
         for _i in range(3):
             f.write(struct.pack("III", 0, 0, 0))
 
-
     def write_rpyc_data(self, f, slot, data):
         """
         Writes data into `slot` of a .rpyc file. The data should be a binary
@@ -524,8 +542,7 @@ class Script(object):
         # Generate translate nodes.
         renpy.translation.restructure(stmts)
 
-
-    def load_file(self, dir, fn): #@ReservedAssignment
+    def load_file(self, dir, fn):  # @ReservedAssignment
 
         if fn.endswith(".rpy") or fn.endswith(".rpym"):
 
@@ -565,27 +582,31 @@ class Script(object):
 
             self.assign_names(stmts, fullfn)
 
-            try:
-                f = file(rpycfn, "wb")
+            if not renpy.macapp:
 
-                self.write_rpyc_header(f)
-                self.write_rpyc_data(f, 1, dumps((data, stmts), 2))
-            except:
-                pass
+                try:
+                    f = file(rpycfn, "wb")
+
+                    self.write_rpyc_header(f)
+                    self.write_rpyc_data(f, 1, dumps((data, stmts), 2))
+                except:
+                    pass
 
             self.static_transforms(stmts)
 
-            try:
-                self.write_rpyc_data(f, 2, dumps((data, stmts), 2))
+            if not renpy.macapp:
 
-                with open(fullfn, "rU") as fullf:
-                    rpydigest = md5.md5(fullf.read()).digest()
+                try:
+                    self.write_rpyc_data(f, 2, dumps((data, stmts), 2))
 
-                self.write_rpyc_md5(f, rpydigest)
+                    with open(fullfn, "rU") as fullf:
+                        rpydigest = hashlib.md5(fullf.read()).digest()
 
-                f.close()
-            except:
-                pass
+                    self.write_rpyc_md5(f, rpydigest)
+
+                    f.close()
+                except:
+                    pass
 
             self.loaded_rpy = True
 
@@ -615,7 +636,7 @@ class Script(object):
                     return None, None
 
                 if data is None:
-                    print "Failed to load", fn
+                    print("Failed to load", fn)
                     return None, None
 
                 if not isinstance(data, dict):
@@ -638,7 +659,7 @@ class Script(object):
 
         return data, stmts
 
-    def load_appropriate_file(self, compiled, source, dir, fn, initcode): #@ReservedAssignment
+    def load_appropriate_file(self, compiled, source, dir, fn, initcode):  # @ReservedAssignment
         # This can only be a .rpyc file, since we're loading it
         # from an archive.
 
@@ -652,8 +673,8 @@ class Script(object):
                 raise Exception("Could not load from archive %s." % (lastfn,))
 
             f = renpy.loader.load(fn + compiled)
-            f.seek(-md5.digest_size, 2)
-            digest = f.read(md5.digest_size)
+            f.seek(-hashlib.md5().digest_size, 2)
+            digest = f.read(hashlib.md5().digest_size)
             f.close()
 
         else:
@@ -667,15 +688,15 @@ class Script(object):
 
             if os.path.exists(rpyfn):
                 with open(rpyfn, "rU") as f:
-                    rpydigest = md5.md5(f.read()).digest()
+                    rpydigest = hashlib.md5(f.read()).digest()
             else:
                 rpydigest = None
 
             try:
                 if os.path.exists(rpycfn):
                     with open(rpycfn, "rb") as f:
-                        f.seek(-md5.digest_size, 2)
-                        rpycdigest = f.read(md5.digest_size)
+                        f.seek(-hashlib.md5().digest_size, 2)
+                        rpycdigest = f.read(hashlib.md5().digest_size)
                 else:
                     rpycdigest = None
             except:
@@ -684,7 +705,7 @@ class Script(object):
             if os.path.exists(rpyfn) and os.path.exists(rpycfn):
 
                 # Are we forcing a compile?
-                force_compile = renpy.game.args.compile # @UndefinedVariable
+                force_compile = renpy.game.args.compile  # @UndefinedVariable
 
                 # Use the source file here since it'll be loaded if it exists.
                 lastfn = rpyfn
@@ -698,11 +719,11 @@ class Script(object):
                         data, stmts = self.load_file(dir, fn + compiled)
 
                         if data is None:
-                            print "Could not load " + rpycfn
+                            print("Could not load " + rpycfn)
 
                 except:
                     if "RENPY_RPYC_EXCEPTIONS" in os.environ:
-                        print "While loading", rpycfn
+                        print("While loading", rpycfn)
                         raise
 
                     pass
@@ -735,11 +756,10 @@ class Script(object):
         elif self.key != data['key']:
             raise Exception( fn + " does not share a key with at least one .rpyc file. To fix, delete all .rpyc files, or rerun Ren'Py with the --lock option.")
 
-        self.finish_load(stmts, initcode, filename=fn + source)
+        self.finish_load(stmts, initcode, filename=lastfn)
 
         self.digest.update(digest)
 
-
     def init_bytecode(self):
         """
         Init/Loads the bytecode cache.
@@ -750,6 +770,7 @@ class Script(object):
             version, cache = loads(renpy.loader.load(BYTECODE_FILE).read().decode("zlib"))
             if version == BYTECODE_VERSION:
                 self.bytecode_oldcache = cache
+
         except:
             pass
 
@@ -759,6 +780,14 @@ class Script(object):
         cache. Clears out self.all_pycode.
         """
 
+        for i in self.all_pyexpr:
+            try:
+                renpy.python.py_compile(i, 'eval')
+            except:
+                pass
+
+        self.all_pyexpr = [ ]
+
         # Update all of the PyCode objects in the system with the loaded
         # bytecode.
         for i in self.all_pycode:
@@ -778,10 +807,12 @@ class Script(object):
 
                     if i.mode == 'exec':
                         code = renpy.python.py_compile_exec_bytecode(i.source, filename=i.location[0], lineno=i.location[1])
+                    elif i.mode == 'hide':
+                        code = renpy.python.py_compile_hide_bytecode(i.source, filename=i.location[0], lineno=i.location[1])
                     elif i.mode == 'eval':
                         code = renpy.python.py_compile_eval_bytecode(i.source, filename=i.location[0], lineno=i.location[1])
 
-                except SyntaxError, e:
+                except SyntaxError as e:
 
                     text = e.text
 
@@ -794,11 +825,11 @@ class Script(object):
                         text = text.decode("latin-1")
 
                     pem = renpy.parser.ParseError(
-                        filename = e.filename,
-                        number = e.lineno,
-                        msg = e.msg,
-                        line = text,
-                        pos = e.offset)
+                        filename=e.filename,
+                        number=e.lineno,
+                        msg=e.msg,
+                        line=text,
+                        pos=e.offset)
 
                     renpy.parser.parse_errors.append(pem.message)
 
@@ -811,8 +842,10 @@ class Script(object):
 
         self.all_pycode = [ ]
 
-
     def save_bytecode(self):
+        if renpy.macapp:
+            return
+
         if self.bytecode_dirty:
             try:
                 fn = renpy.loader.get_path(BYTECODE_FILE)
@@ -823,7 +856,6 @@ class Script(object):
             except:
                 pass
 
-
     def lookup(self, label):
         """
         Looks up the given label in the game. If the label is not found,
@@ -862,3 +894,15 @@ class Script(object):
             i.analyze()
 
         self.need_analysis = [ ]
+
+    def report_duplicate_labels(self):
+        if not renpy.config.developer:
+            return
+
+        if renpy.config.ignore_duplicate_labels:
+            return
+
+        renpy.parser.parse_errors = self.duplicate_labels
+
+        if renpy.parser.report_parse_errors():
+            raise SystemExit(-1)
