@@ -179,7 +179,7 @@ class ScriptTranslator(object):
             tl = self.language_translates.get((identifier, language), None)
 
             if (tl is None) and alternate:
-                tl = self.language_translates.get((identifier, language), alternate)
+                tl = self.language_translates.get((identifier, language), None)
 
         else:
             tl = None
@@ -369,7 +369,15 @@ class StringTranslator(object):
 
     def add(self, old, new, newloc):
         if old in self.translations:
-            raise Exception("A translation for %r already exists." % old)
+
+            if old in self.translation_loc:
+                print(newloc, self.translation_loc[old])
+                fn, line = self.translation_loc[old]
+                raise Exception("A translation for \"{}\" already exists at {}:{}.".format(
+                    quote_unicode(old), fn, line))
+            else:
+                raise Exception("A translation for \"{}\" already exists.".format(
+                    quote_unicode(old)))
 
         self.translations[old] = new
 
@@ -436,8 +444,12 @@ Default = renpy.object.Sentinel("default")
 
 def translate_string(s, language=Default):
     """
-    Translates interface string `s` to `language`. If `languages` is Default,
-    uses the language set in the preferences.
+    :doc: translate_string
+    :name: renpy.translate_string
+
+    Translates interface string `s` to `language`. If `language` is Default,
+    uses the language set in the preferences. This does not mark `s` to be
+    translated.
     """
 
     if language is Default:
@@ -565,8 +577,6 @@ def old_change_language(tl, language):
 
 def new_change_language(tl, language):
 
-    renpy.config.init_system_styles()
-
     for i in tl.python[language]:
         renpy.python.py_exec_bytecode(i.code.bytecode)
 
@@ -588,6 +598,8 @@ def new_change_language(tl, language):
 
     renpy.game.invoke_in_new_context(run_blocks)
 
+    renpy.config.init_system_styles()
+
 
 def change_language(language, force=False):
     """
